import numpy as np

from .activation import sigmoid
from .module import Module
from ..initializations import init
from ..values import Value

__all__ = ["RNN", "LSTM"]


class _BaseRNN(Module):
    def __init__(self, n_input: int, n_output: int, bias: bool = True, nonlinearity: str = "tanh"):
        """
        :param n_input: int - size of each input sample.
        :param n_output: int - size of each output sample.
        :param bias: bool - consider bias in layer computation or not.
        :param nonlinearity: str - which non-linearity to use (currently not supported).
        """
        super(_BaseRNN, self).__init__()
        self.n_input = n_input
        self.n_output = n_output

        self.weight_xh: Value = init.kaiming_uniform(shape=(n_input, n_output), nonlinearity=nonlinearity)
        self.weight_hh: Value = init.kaiming_uniform(shape=(n_output, n_output), nonlinearity=nonlinearity)
        self.bias: Value | None = init.kaiming_uniform(shape=(n_output,), nonlinearity=nonlinearity) if bias else None

        self.hidden_states = None
        self.x = None


class RNN(_BaseRNN):
    """
    Vanilla recurrent neural network.
    """

    def forward(self, x: np.ndarray, h0: np.ndarray = None) -> (np.ndarray, np.ndarray):
        """
        Forward pass for RNN in a 'for' loop way.

        :param x: numpy array (batch_size, sequence_length, input_size) - incoming data.
        :param h0: numpy array (bach_size, hidden_size) - initial hidden state for the input sequence.
        :return: tuple of two numpy arrays (output, hn):
                output: numpy array (batch_size, sequence_length, hidden_size) - hidden states for all time steps.
                hn: numpy array (batch_size, hidden_size) - final hidden state.
        """
        self.x = x.copy()
        batch_size, sequence_length, _ = x.shape

        h = np.copy(h0) if h0 is not None else np.zeros(shape=(batch_size, self.n_output))
        self.hidden_states = np.zeros(shape=(batch_size, sequence_length + 1, self.n_output))
        self.hidden_states[:, 0, :] = h

        for idx in range(sequence_length):
            h_raw = np.dot(self.x[:, idx, :], self.weight_xh.data) + np.dot(h, self.weight_hh.data)

            if self.bias is not None:
                h_raw += self.bias.data

            self.hidden_states[:, idx + 1, :] = np.tanh(h_raw)
            h = self.hidden_states[:, idx + 1, :]

        return self.hidden_states[:, 1:, :], self.hidden_states[:, -1, :]

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to incoming data.

        :param d_out: numpy array (batch_size, sequence_length, hidden_size) -
        gradient of loss function with respect to output of forward pass.

        :return: numpy array (batch_size, sequence_length, input_size) -
        gradient with respect to x, the same shape as d_out.
        """
        dx = np.zeros_like(self.x)
        _, sequence_length, _ = dx.shape

        dh_next = np.zeros_like(self.hidden_states[:, 0, :])
        for idx in reversed(range(sequence_length)):
            dh_curr = d_out[:, idx, :] + dh_next
            dh_raw = (1 - np.square(self.hidden_states[:, idx + 1, :])) * dh_curr

            if self.bias is not None:
                self.bias.grad += dh_raw.sum(axis=0)
            self.weight_xh.grad += np.dot(self.x[:, idx, :].T, dh_raw)

            self.weight_hh.grad += np.dot(self.hidden_states[:, idx, :].T, dh_raw)
            dx[:, idx, :] = np.dot(dh_raw, self.weight_xh.data.T)

            dh_next = np.dot(dh_raw, self.weight_hh.data.T)

        return dx


class LSTM(_BaseRNN):
    """
    Vanilla LSTM neural network.
    """
    def __init__(self, n_input: int, n_output: int, bias: bool = True, nonlinearity: str = "tanh"):
        """
        :param n_input: int - size of each input sample.
        :param n_output: int - size of each output sample.
        :param bias: bool - consider bias in layer computation or not.
        :param nonlinearity: str - which non-linearity to use (currently not supported).
        """
        super(LSTM, self).__init__(n_input, n_output, bias, nonlinearity)

        self.cell_states = None

        self.weight_xh: Value = init.kaiming_uniform(shape=(n_input, 4 * n_output), nonlinearity=nonlinearity)
        self.weight_hh: Value = init.kaiming_uniform(shape=(n_output, 4 * n_output), nonlinearity=nonlinearity)
        self.bias = init.kaiming_uniform(shape=(4 * n_output,), nonlinearity=nonlinearity) if bias else None

        self.block_input = None
        self.output_gates = None
        self.forget_gates = None
        self.input_gates = None

    def forward(self, x: np.ndarray, h0: np.ndarray = None, c0: np.ndarray = None) -> (np.ndarray, np.ndarray):
        """
        Forward pass for LSTM neural network in a 'for' loop way.

        :param x: numpy array (batch_size, sequence_length, input_size) - incoming data.
        :param h0: numpy array (bach_size, hidden_size) - initial hidden state for the input sequence.
        :param c0:  numpy array (batch_size, hidden_size) - initial cell state for the input sequence.

        :return: tuple of two numpy arrays (output, hn):
                output: numpy array (batch_size, sequence_length, hidden_size) - hidden states for all time steps.
                hn: numpy array (batch_size, hidden_size) - final hidden state.
        """
        self.x = x.copy()
        batch_size, sequence_length, _ = x.shape

        h = np.copy(h0) if h0 is not None else np.zeros(shape=(batch_size, self.n_output))
        c = np.copy(c0) if c0 is not None else np.zeros(shape=(batch_size, self.n_output))

        self.hidden_states = np.zeros(shape=(batch_size, sequence_length + 1, self.n_output))
        self.hidden_states[:, 0, :] = h

        self.cell_states = np.zeros(shape=(batch_size, sequence_length + 1, self.n_output))
        self.cell_states[:, 0, :] = c

        self.input_gates = np.zeros(shape=(batch_size, sequence_length, self.n_output))
        self.forget_gates = np.zeros(shape=(batch_size, sequence_length, self.n_output))
        self.output_gates = np.zeros(shape=(batch_size, sequence_length, self.n_output))
        self.block_input = np.zeros(shape=(batch_size, sequence_length, self.n_output))

        for idx in range(sequence_length):
            a_raw = np.dot(h, self.weight_hh.data) + np.dot(self.x[:, idx, :], self.weight_xh.data)
            if self.bias is not None:
                a_raw += self.bias.data

            ai, af, ao, ag = np.array_split(a_raw, 4, axis=1)

            self.input_gates[:, idx, :] = sigmoid(ai)
            self.forget_gates[:, idx, :] = sigmoid(af)
            self.output_gates[:, idx, :] = sigmoid(ao)
            self.block_input[:, idx, :] = np.tanh(ag)

            c = self.forget_gates[:, idx, :] * c + self.input_gates[:, idx, :] * self.block_input[:, idx, :]
            h = self.output_gates[:, idx, :] * np.tanh(c)

            self.cell_states[:, idx + 1, :] = c
            self.hidden_states[:, idx + 1, :] = h

        return self.hidden_states[:, 1:, :], self.hidden_states[:, -1, :]

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to incoming data.

        :param d_out: numpy array (batch_size, sequence_length, hidden_size) -
        gradient of loss function with respect to output of forward pass.

        :return: numpy array (batch_size, sequence_length, input_size) -
        gradient with respect to x, the same shape as d_out.
        """
        dx = np.zeros_like(self.x)

        _, sequence_length, _ = dx.shape

        dh_next = np.zeros_like(self.hidden_states[:, 0, :])
        dc_next = np.zeros_like(self.cell_states[:, 0, :])

        for idx in reversed(range(sequence_length)):
            dh_curr = d_out[:, idx, :] + dh_next
            dc_curr = self.output_gates[:, idx, :] * np.cosh(self.cell_states[:, idx + 1, :])**(-2) * dh_curr + dc_next

            doutput_gate = np.tanh(self.cell_states[:, idx + 1, :]) * dh_curr
            dforget_gate = self.cell_states[:, idx, :] * dc_curr

            dc_next = self.forget_gates[:, idx, :] * dc_curr

            dinput_gate = self.block_input[:, idx, :] * dc_curr
            dblock_input = self.input_gates[:, idx, :] * dc_curr

            d_ag = (1 - self.block_input[:, idx, :] ** 2) * dblock_input
            d_ao = (self.output_gates[:, idx, :] * (1 - self.output_gates[:, idx, :])) * doutput_gate
            d_af = (self.forget_gates[:, idx, :] * (1 - self.forget_gates[:, idx, :])) * dforget_gate
            d_ai = (self.input_gates[:, idx, :] * (1 - self.input_gates[:, idx, :])) * dinput_gate

            da = np.concatenate((d_ai, d_af, d_ao, d_ag), axis=1)

            self.weight_xh.grad += np.dot(self.x[:, idx, :].transpose(), da)
            self.weight_hh.grad += np.dot(self.hidden_states[:, idx, :].transpose(), da)
            if self.bias is not None:
                self.bias.grad += da.sum(axis=0)

            dx[:, idx, :] += np.dot(da, self.weight_xh.data.transpose())
            dh_prev = np.dot(da, self.weight_hh.data.transpose())

            dh_next = dh_prev

        return dx



import numpy as np


class Value:
    """
    Trainable parameter
    Contain both data and gradient values
    """

    def __init__(self, value: np.ndarray):
        """
        :param value: numpy array (n-dimensional) - layer parameters
        """
        self.data = value
        self.grad = np.zeros_like(value)

    def __repr__(self):
        return f"Value(data={self.data}, grad={self.grad})"



import copy

import numpy as np

from .module import Module
from ..values import Value


class Sequential(Module):
    """
    Module for structuring neural network layers
    """

    def __init__(self, *args: Module):
        """
        :param args: Layer - collection of neural network layers that will be executed sequentially
        """
        super(Sequential, self).__init__()

        self._layers: list[Module] = []

        for arg in args:
            if not isinstance(arg, Module):
                print(type(arg))
                raise Exception("Invalid argument for layer")

            self._layers.append(arg)

    def __call__(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Iterates over a list of neural network layers and
        sequentially performs a forward pass depending on the mode state

        :param x: numpy array (n-dimensional) - data for prediction
        :return: numpy array (n-dimensional) - logits after performing forward pass
        """
        copy_X = copy.deepcopy(x)
        for layer in self._layers:
            copy_X = layer(copy_X)

        return copy_X

    def backward(self, d_out: np.ndarray):
        """
        Iterates over a reversed list of neural network layers
        and calculates backward pass for each of these layers

        :param d_out: numpy array (n-dimensional) - gradient of loss function
        """
        d_out_copy = copy.deepcopy(d_out)
        for layer in reversed(self._layers):
            d_out_copy = layer.backward(d_out_copy)

        return d_out_copy

    def eval(self):
        """
        Iterates over a list of neural network layers and sets their mode to evaluation.
        """
        for layer in self._layers:
            layer.eval()

    def train(self):
        """
        Iterates over a list of neural network layers and sets their mode to training.
        """
        for layer in self._layers:
            layer.train()

    @property
    def params(self) -> dict[str, Value]:
        """
        Passes through every layer of the neural network and gathers their
        parameters into dictionary

        :return: dict (param_name: str, param: Param)
        """

        for idx, layer in enumerate(self._layers):
            for param_name, param in layer.params.items():
                self._params[f"{param_name}_{idx}"] = param

        return self._params



from abc import ABC

import numpy as np

from torchy.nn.values import Value


class _Optim(ABC):
    """
    Abstract class for all gradient optimization techniques.
    """

    def __init__(self, params: dict[str, Value], lr: float = 1e-4, weight_decay: float = 0.0):
        """
        :param params: dict (parameter_name, parameter) - contains all neural network parameters wrapped
        into dictionary where key is parameter name in a specific layer and values is weight/bias with its own
        data and gradient of that specific layer.
        :param lr: float - learning rate of a neural network model.
        :param weight_decay: float - regularization parameter.
        """
        self._params = params
        self.lr = lr
        self.weight_decay = weight_decay / 2

    def zero_grad(self):
        """
        Goes through the self.params and zeroes parameter gradients.
        """
        for param in self._params.values():
            param.grad = np.zeros_like(param.grad)

    def _update(self, param_name: str, param: Value):
        """
        Single parameter update.

        :param param_name: str - parameter name.
        :param param: Value - holds parameter state (data and gradient).
        """
        pass

    def step(self):
        """
        Loops over parameter dictionary, unpacks it by key and values and then passes
        key and values to the update method.
        """
        for param_name, param in self._params.items():
            self._update(param_name, param)


class SGD(_Optim):
    """
    Implementation of classic Stochastic Gradient Descent.
    """

    def _update(self, param_name: str, param: Value):
        """
        Single SGD parameter update with regularization.

        :param param_name: str - parameter name.
        :param param: Value - holds parameter state (data and gradient).
        """
        gradient = param.grad + self.weight_decay * param.data
        param.data -= self.lr * gradient


class MomentumSGD(_Optim):
    """
    Stochastic Gradient Descent with Momentum
    """

    def __init__(self, params: dict[str, Value], lr: float = 1e-4, weight_decay: float = 0.0, momentum: float = 0.9):
        """
        :param params: dict (parameter_name: str, parameter: Value) - contains all neural network parameters wrapped
        into dictionary where key is parameter name in a specific layer and values is weight/bias with its own
        data and gradient of that specific layer.
        :param lr: float - learning rate of a neural network model.
        :param weight_decay: float - regularization parameter.
        :param momentum: float (0.9-0.999) - momentum parameter (default = 0.9).
        """
        super().__init__(params=params, lr=lr, weight_decay=weight_decay)
        self.momentum = momentum
        self._velocities = {}

    def _update(self, param_name: str, param: Value):
        """
        Single MomentumSGD parameter update with regularization.

        :param param_name: str - parameter name.
        :param param: Value - holds parameter state (data and gradient).
        """
        velocity = self._velocities.get(param_name, np.zeros_like(param.data))
        gradient = param.grad + self.weight_decay * param.data

        self._velocities[param_name] = self.momentum * velocity - self.lr * gradient
        param.data += self._velocities[param_name]


class Adagrad(_Optim):
    """
    Adagrad neural network optimization.
    """

    def __init__(self, params: dict[str, Value], lr: float = 1e-2, weight_decay: float = 0.0, eps: float = 1e-8):
        """
        :param params: dict (parameter_name, parameter) - contains all neural network parameters wrapped
        into dictionary where key is parameter name in a specific layer and values is weight/bias with its own
        data and gradient of that specific layer.
        :param lr: float - learning rate of a neural network model.
        :param weight_decay: float - regularization parameter.
        :param eps: float - values added to numerical stability in the denominator.
        """
        super().__init__(params=params, lr=lr, weight_decay=weight_decay)
        self.eps = eps
        self._accumulated = {}

    def _update(self, param_name: str, param: Value):
        """
        Single Adagrad parameter update with regularization.

        :param param_name: str - parameter name.
        :param param: Value - holds parameter state (data and gradient).
        """
        grad_squared = np.square(param.grad)
        grad_accumulated = self._accumulated.get(param_name, np.zeros_like(param.grad))
        self._accumulated[param_name] = grad_accumulated + grad_squared

        adaptive_lr = self.lr / np.sqrt(self._accumulated[param_name] + self.eps)

        param.data -= (adaptive_lr * param.grad) + (param.data * self.weight_decay)


class RMSProp(_Optim):
    """
    RMSProp gradient optimization algorithm.
    """

    def __init__(self, params: dict[str, Value], lr: float = 1e-2, weight_decay: float = 0.0, rho: float = 0.99,
                 eps: float = 1e-8):
        """
        :param params: dict (parameter_name, parameter) - contains all neural network parameters wrapped.
        into dictionary where key is parameter name in a specific layer and values is weight/bias with its own
        data and gradient of that specific layer.
        :param lr: float - learning rate of a neural network model.
        :param weight_decay: float - regularization parameter.
        :param rho: float - momentum factor.
        :param eps: float - values added to numerical stability in the denominator.
        """
        super().__init__(params=params, lr=lr, weight_decay=weight_decay)
        self.rho = rho
        self.eps = eps
        self._accumulated = {}

    def _update(self, param_name: str, param: Value):
        """
        Single RMRSProp parameter update with regularization.

        :param param_name: str - parameter name.
        :param param: Value - holds parameter state (data and gradient).
        """
        grad_squared = np.square(param.grad)
        grad_accumulated = self._accumulated.get(param_name, np.zeros_like(param.grad))
        self._accumulated[param_name] = self.rho * grad_accumulated + (1 - self.rho) * grad_squared

        adaptive_lr = self.lr / np.sqrt(self._accumulated[param_name] + self.eps)

        param.data -= (adaptive_lr * param.grad) + (param.data * self.weight_decay)


class Adam(_Optim):
    """
    Adam gradient optimization algorithm.
    """

    def __init__(self,
                 params: dict[str, Value],
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 lr: float = 1e-2,
                 weight_decay: float = 0.0,
                 eps: float = 1e-8,
                 t: int = 0
                 ):
        """
        :param params: dict (parameter_name, parameter) - contains all neural network parameters wrapped
        into dictionary where key is parameter name in a specific layer and values is weight/bias with its own
        data and gradient of that specific layer.
        :param beta1: float - coefficient used for computing running averages.
        :param beta2: float - coefficient used for computing running averages.
        :param lr: float - learning rate of a neural network model.
        :param weight_decay: float - regularization parameter.
        :param eps: float - values added to numerical stability in the denominator.
        :param t: int - iteration number for warming up Adam.
        """

        super().__init__(params=params, lr=lr, weight_decay=weight_decay)

        self._velocities = {}
        self._accumulated = {}
        self.eps = eps
        self.beta1 = beta1
        self.beta2 = beta2
        self.t = t

    def _update(self, param_name: str, param: Value):
        """
        Single Adam parameter update with regularization.

        :param param_name: str - parameter name.
        :param param: Value - holds parameter state (data and gradient).
        """
        self.t += 1

        velocity = self._velocities.get(param_name, np.zeros_like(param.data))
        self._velocities[param_name] = self.beta1 * velocity + (1 - self.beta1) * param.grad
        vt = self._velocities[param_name] / (1 - self.beta1 ** self.t)

        accumulated = self._accumulated.get(param_name, np.zeros_like(param.data))
        self._accumulated[param_name] = self.beta2 * accumulated + (1 - self.beta2) * param.grad ** 2
        at = self._accumulated[param_name] / (1 - self.beta2 ** self.t)

        param.data -= self.lr * vt / (np.sqrt(at) + self.eps) + (self.weight_decay * param.data)



import numpy as np

from torchy.nn.initializations import init
from torchy.nn.values import Value
from .module import Module


class Linear(Module):
    """
    Fully-connected / Dense layer
    """

    def __init__(self, n_input: int, n_output: int, bias: bool = True):
        """
        :param n_input: int - size of each input sample.
        :param n_output: int - size of each output sample.
        :param bias: bool - consider bias in layer computation or not
        """
        super(Linear, self).__init__()

        self.weight: Value = init.kaiming_uniform(shape=(n_input, n_output))
        self.bias: Value | None = init.kaiming_uniform(shape=(n_output,)) if bias else None
        self.X = None
        self.out = None

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Computes forward pass for linear layer.

        :param x: numpy array (batch_size, n_input) - incoming data.
        :return: numpy array (batch_size, n_output) - incoming data after linear transformation.
        """
        self.X = np.copy(x)
        self.out = np.dot(self.X, self.weight.data)

        if self.bias is not None:
            self.out += self.bias.data

        return self.out

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes gradient with respect to input, weight and bias

        :param d_out: numpy array (batch_size, n_output) - gradient of loss with respect to an output.
        :return: numpy array (batch_size, n_input) - gradient with respect to input.
        """
        self.weight.grad = np.dot(self.X.T, d_out)

        if self.bias is not None:
            self.bias.grad = np.sum(d_out, axis=0).T

        d_pred = np.dot(d_out, self.weight.data.T)

        return d_pred



from .activation import *
from .batchnorm import *
from .conv import *
from .dense import *
from .dropout import *
from .embedding import *
from .loss import *
from .module import *
from .pooling import *
from .rnn import *
from .sequential import *



import numpy as np

from .module import Module


class Dropout(Module):
    """
    Inverted dropout
    """

    def __init__(self, p: float = 0.5, seed: int = None):
        """
        :param p: float - probability of keeping each neuron.
        :param seed: int - parameter for data representativity (default None).
        """
        super(Dropout, self).__init__()
        self.p = p
        self.seed = np.random.seed(seed) if seed else None

        self.mask = None

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Performs forward pass for inverted dropout.

        :param x: numpy array (n-dimensional) - incoming data.
        :return: numpy array (n-dimensional) - output data, same shape as incoming data
        """
        out = x.copy()
        if self._train:
            self.mask = (np.random.rand(*x.shape) < self.p) / self.p
            out *= self.mask

        return out

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to forward pass input.

        :param d_out: numpy array (n-dimensional) - gradient of loss with respect to output of forward pass
        :return: numpy array (n-dinemsional) - gradient with respect to input, same shape as forward pass input
        """

        return d_out * self.mask



import numpy as np

from .module import Module
from torchy.nn.values import Value


class _BatchNorm(Module):
    """
    Base-like class for batch normalization
    """

    def __init__(self,
                 n_output: int,
                 eps: float = 1e-5,
                 momentum: float = 0.9):
        """
        :param n_output: int - number of output parameters.
        :param eps: float - values added to numerical stability in denominator.
        :param momentum: float - coefficient for computing running mean and variance
        """
        super(_BatchNorm, self).__init__()

        self._out = None
        self._X_norm = None
        self._X_var = None
        self._X_mean = None
        self.x = None

        self.eps = eps
        self.gamma = Value(np.ones(n_output))
        self.beta = Value(np.zeros(n_output))
        self.momentum = momentum

        self.running_mean = np.zeros(n_output)
        self.running_var = np.zeros(n_output)

    def _forward_1d(self, x: np.ndarray) -> np.ndarray:
        """
        Forward pass of batch normalization layer.

        :param x: numpy array (batch_size, n_output) - incoming data.
        :return: numpy array (batch_size, n_output) - result of batchnorm processing.
        """
        if not self._train:
            eval_norm = (x - self.running_mean) / np.sqrt(self.running_var + self.eps)
            self._out = self.gamma.data * eval_norm + self.beta.data
            return self._out

        self.x = np.copy(x)

        self._X_mean = np.mean(x, axis=0, keepdims=True)
        self._X_var = np.var(x, axis=0, keepdims=True)
        self._X_norm = (x - self._X_mean) / np.sqrt(self._X_var + self.eps)

        self._out = self.gamma.data * self._X_norm + self.beta.data

        self.running_mean = self.momentum * self.running_mean + (1 - self.momentum) * self._X_mean
        self.running_var = self.momentum * self.running_var + (1 - self.momentum) * self._X_var

        return self._out

    def _backward_1d(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to x, gamma and beta.

        :param d_out: numpy array (batch_size, n_output) - gradient of loss function with respect to forward pass.
        :return: numpy array (batch_size, n_output) - gradient with respect to self.x.
        """
        self.gamma.grad = (self._X_norm * d_out).sum(axis=0)
        self.beta.grad = d_out.sum(0)

        batch_size = self.x.shape[0]

        sqrt_var_eps = np.sqrt(self._X_var + self.eps)
        dxhat = d_out * self.gamma.data
        dvar = np.sum(dxhat * (self.x - self._X_mean), axis=0) * (-1 / 2) * (self._X_var + self.eps) ** (-3 / 2)
        dmu = np.sum(dxhat * (-1 / sqrt_var_eps), axis=0) + dvar * (-2 / batch_size) * np.sum(self.x - self._X_mean,
                                                                                              axis=0)
        dx = dxhat * (1 / sqrt_var_eps) + dvar * (2 / batch_size) * (self.x - self._X_mean) + dmu / batch_size
        return dx


class BatchNorm1d(_BatchNorm):
    """
    Batch Normalization for one-dimensional layers.
    """

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Forward pass of BatchNorm1d layer.

        :param x: numpy array (batch_size, n_output) - incoming data.
        :return: numpy array (batch_size, n_output) - result of batchnorm processing.
        """
        return self._forward_1d(x)

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to x, gamma and beta.

        :param d_out: numpy array (batch_size, n_output) - gradient of loss function with respect to forward pass.
        :return: numpy array (batch_size, n_output) - gradient with respect to self.x.
        """
        return self._backward_1d(d_out)


class BatchNorm2d(_BatchNorm):
    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Forward pass of BatchNorm2d layer.

        Transforms input data to a two-dimensional array and basically calculates
        batch normalization in one-dimensional representation via _forward_ndim().

        After performing calculation, transforms output data to appropriate shape.

        :param x: numpy array (batch_size, in_channels, height, width) - incoming data.
        :return: numpy array (batch_size, in_channels, height, width) - result of batchnorm processing.
        """
        batch_size, in_channels, height, width = x.shape
        x_reshaped = x.transpose((0, 2, 3, 1)).reshape((-1, in_channels))
        out = self._forward_1d(x_reshaped)
        return out.reshape((batch_size, height, width, in_channels)).transpose((0, 3, 1, 2))

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to x, gamma and beta.

        In terms of shape, d_out acts the same as x in forward pass.

        :param d_out: numpy array (batch_size, n_output) - gradient of loss function with respect to forward pass.
        :return: numpy array (batch_size, n_output) - gradient with respect to self.x.
        """
        batch_size, in_channels, height, width = d_out.shape
        d_out_reshaped = d_out.transpose((0, 2, 3, 1)).reshape((-1, in_channels))
        dx = self._backward_1d(d_out_reshaped)
        return dx.reshape((batch_size, height, width, in_channels)).transpose((0, 3, 1, 2))



import numpy as np

from .module import Module


class MaxPool2d(Module):
    """
    Max pooling layer for 2-dimensional feature map (Convolutional Layer).
    """

    def __init__(self, kernel_size: int, stride: int = None, padding: int = 0):
        """
        :param kernel_size: int - size of max pooling kernel.
        :param stride: int - stride of max pooling  kernel (default = kernel_size).
        :param padding: int = padding added to all axis with respect to input (default = 0).
        """
        super(MaxPool2d, self).__init__()

        self.X = None
        self.kernel_size = kernel_size
        self.stride = stride if stride is not None else kernel_size
        self.padding = padding
        self._padding_width = ((0, 0), (0, 0), (padding, padding), (padding, padding))

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Forward pass of max pooling layer.

        :param x: numpy array (batch_size, in_channels, height, width) - data to perform max pooling.
        :return: numpy array (batch_size, in_channels, out_height, out_width) - result of max pooling 'x'.
        """
        self.X = np.pad(x, pad_width=self._padding_width, mode="constant", constant_values=0)
        batch_size, in_channels, height, width = self.X.shape

        out_height = (height + 2 * self.padding - self.kernel_size) // self.stride + 1
        out_width = (width + 2 * self.padding - self.kernel_size) // self.stride + 1

        out = np.zeros(shape=(batch_size, in_channels, out_height, out_width))
        for oh in range(out_height):
            for ow in range(out_width):
                oh_step = self.stride * oh
                ow_step = self.stride * ow

                input_region = self.X[:, :, oh_step:self.kernel_size + oh_step, ow_step:self.kernel_size + ow_step]
                out[:, :, oh, ow] += np.max(input_region, axis=(2, 3))

        return out

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to self.x.

        :param d_out: numpy array (batch_size, in_channels, out_height, out_width) - gradient of loss function with
        respect to output of forward pass.
        :return: numpy array (batch_size, in_channels, height, width) - gradient with respect to self.x.
        """
        _, in_channels, out_height, out_width = d_out.shape

        d_pred = np.zeros_like(self.X)
        for oh in range(out_height):
            for ow in range(out_width):
                oh_step = self.stride * oh
                ow_step = self.stride * ow

                output_region = self.X[:, :, oh_step:oh_step + self.kernel_size, ow_step:ow_step + self.kernel_size]
                grad = d_out[:, :, oh, ow][:, :, np.newaxis, np.newaxis]
                mask = (output_region == np.max(output_region, (2, 3))[:, :, np.newaxis, np.newaxis])
                d_pred[:, :, oh_step:oh_step + self.kernel_size, ow_step:ow_step + self.kernel_size] += grad * mask

        return d_pred



import numpy as np


def compute_softmax(y_pred: np.ndarray) -> np.ndarray:
    """
    Computes softmax function and subtracts the maximum values from each row for calculation stability.

    :param y_pred: numpy array (batch_size, n_output) - predictions computed by neural network with a range
    of values from (-infinity, +infinity).
    :return: numpy array (batch_size, n_output) - result of softmax function with a range of values from (0, 1)
    with a total row sum of 1.
    """
    exps = np.exp(y_pred - np.max(y_pred, axis=1, keepdims=True))
    return exps / np.sum(exps, axis=1, keepdims=True)


class CrossEntropyLoss:
    """
    Cross-entropy Loss

    Computes the cross-entropy loss between input logits and the target
    """

    def __call__(self, y_pred: np.ndarray, y_true: np.ndarray) -> (float, np.ndarray):
        """

        :param y_pred: numpy array (batch_size, n_output) - predictions computed by neural network with a range
        of values from (-infinity, +infinity).
        :param y_true: numpy array (batch_size) - indices of ground truth values.
        :return: loss (float) - cross-entropy loss.
        :return grad (batch_size, n_output) - gradient of loss function with respect to softmax.
        """
        batch_size = y_true.shape[0]
        num_classes = y_pred.shape[1]

        y_true_one_hot = np.identity(num_classes)[y_true.reshape(-1)]
        logits = compute_softmax(y_pred)

        loss = -np.sum(y_true_one_hot * np.log(logits + 1e-10)) / batch_size

        grad = (logits - y_true_one_hot) / batch_size

        return loss, grad



import numpy as np

from torchy.nn.initializations import init
from torchy.nn.values import Value
from .module import Module


class Embedding(Module):
    """
    Word embedding layer.
    """

    def __init__(self, num_embeddings: int, embedding_dim: int):
        """
        :param num_embeddings: int - dictionary size.
        :param embedding_dim: int - embedding dimension for each word in dictionary.
        """
        super(Embedding, self).__init__()

        self.weight = init.normal(shape=(num_embeddings, embedding_dim))
        self.x = None

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Computes forward pass for embedding layer.

        :param x: numpy array (batch_size, dictionary_size) - incoming data.
        :return: numpy array (batch_size, sequence_length, embedding_dim) - word embeddings.
        """
        self.x = np.copy(x)

        return self.weight.data[self.x]

    def backward(self, d_out: np.ndarray) -> None:
        """
        Computes backward pass with respect to incoming data.

        :param d_out: numpy array (batch_size, sequence_length, embedding_dim) -
        gradient of loss function with respect to output of forward pass.
        """
        np.add.at(self.weight.grad, self.x, d_out)

    @classmethod
    def from_pretrained(cls, weight: np.ndarray) -> 'Embedding':
        """
        Takes pretrained weight parameters and creates new instance of an embedding object
        with pretrained weights.

        :param weight: numpy array (num_embeddings, embedding_dim) - pretrained embedding weights.
        :return: Embedding - embedding class instance with pretrained weights.
        """
        num_embeddings, embedding_dim = weight.shape
        embedding = cls(num_embeddings, embedding_dim)
        embedding.weight = Value(weight)

        return embedding



import numpy as np

from torchy.nn.values import Value


def kaiming_uniform(shape: tuple[int, ...], nonlinearity: str = "relu") -> Value:
    """
    Kaiming parameter initializations with uniform distribution.

    :param shape: Tuple[int] - parameter shape.
    :param nonlinearity: str - non=linearity for proper gain calculation.
    :return: Value - initialized module parameter.
    """
    gain = _calculate_gain(nonlinearity)
    fan_in = _calculate_fan_in(shape)
    std = gain / np.sqrt(fan_in)
    bound = std * np.sqrt(3.0)
    return Value(np.random.uniform(-bound, bound, size=shape))


def normal(shape: tuple[int, ...]) -> Value:
    """
    Parameter initializations with normal distribution with mean = 0 and std = 1.

    :param shape: Tuple[int] - parameter shape.
    :return: Value - initialized module parameter.
    """
    return Value(np.random.randn(*shape))


def kaiming_normal(shape: tuple[int, ...], nonlinearity: str = "relu") -> Value:
    """
    Kaiming parameter initializations with normal distribution.

    :param shape: Tuple[int] - parameter shape.
    :param nonlinearity: str - nonlinearity for proper gain calculation.
    :return: Value - initialized module parameter.
    """
    gain = _calculate_gain(nonlinearity)
    fan_in = _calculate_fan_in(shape)
    std = gain / np.sqrt(fan_in)
    return Value(std * np.random.randn(*shape))


def _calculate_gain(nonlinearity: str) -> float:
    """
    Takes non-linearity as input and calculates gain for proper parameter initializations.

    :param nonlinearity: str - non-linearity for gain calculation.
    :return: float - calculated gain.
    """
    if nonlinearity == 'relu':
        return np.sqrt(2.0)
    elif nonlinearity == 'sigmoid':
        return 1
    elif nonlinearity == 'tanh':
        return 5.0 / 3
    else:
        raise NotImplementedError(f"Gain calculation for '{nonlinearity}' non-linearity is not implemented.")


def _calculate_fan_in(shape: tuple[int, ...]) -> int:
    """
    Takes shape as input and calculates parameter fan in depending on the parameter shape size.

    :param shape: Tuple[int] - parameter shape.
    :return:int - calculated fan in.
    """
    if len(shape) == 2 or len(shape) == 1:
        return shape[0]
    elif len(shape) == 4:
        return shape[1] * shape[2] * shape[3]
    else:
        raise NotImplementedError(f"Fan in calculation for {shape} shape is not implemented.")



import numpy as np

from .module import Module


class ReLU(Module):
    """
    Rectified Linear Unit activation function.
    """

    def __init__(self):
        super(ReLU, self).__init__()

        self._mask = None

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Forward pass of ReLU layer

        :param x: numpy array (n-dimensional) - incoming data.
        :return: numpy array (n-dimensional) - data after performing activation function on it, same shape
        as x.
        """
        self._mask = x > 0
        return np.where(self._mask, x, 0)

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to incoming data.

        :param d_out: numpy array (n-dimensional) - gradient of loss function with respect to output of forward pass.
        :return: numpy array (n-dimensional) - gradient with respect to x, the same shape as d_out.
        """
        return d_out * self._mask


class Tanh(Module):
    """
    Hyperbolic tangent activation function.
    """

    def __init__(self):
        super(Tanh, self).__init__()
        self.x = None

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Forward pass of Tanh layer

        :param x: numpy array (n-dimensional) - incoming data.
        :return: numpy array (n-dimensional) - data after performing activation function on it, same shape as x.
        """
        self.x = np.tanh(x)
        return self.x

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes backward pass with respect to incoming data.

        :param d_out: numpy array (n-dimensional) - gradient of loss function with respect to output of forward pass.
        :return: numpy array (n-dimensional) - gradient with respect to x, the same shape as d_out.
        """
        return (1 - self.x ** 2) * d_out


def sigmoid(x):
    """A numerically stable version of the logistic sigmoid function."""
    pos_mask = x >= 0
    neg_mask = x < 0
    z = np.zeros_like(x)
    z[pos_mask] = np.exp(-x[pos_mask])
    z[neg_mask] = np.exp(x[neg_mask])
    top = np.ones_like(x)
    top[neg_mask] = z[neg_mask]
    return top / (1 + z)


from abc import ABC

import numpy as np

from .optim import _Optim


class _Scheduler(ABC):
    """
    Abstract class for scheduler algorithm.
    """

    def __init__(self, optimizer: _Optim, verbose: bool = False):
        """
        :param optimizer: _Optim - gradient descent optimizer.
        :param verbose: bool - on/off logging every scheduler step.
        """
        self.optimizer = optimizer
        self.verbose = verbose

    def step(self, *args, **kwargs):
        """
        Abstract method for scheduler step.

        :param args:
        :param kwargs:
        """
        pass


class StepLR(_Scheduler):
    """
    Decays learning rate of optimizer every self.step_size by self.gamma.
    """

    def __init__(self, optimizer: _Optim, step_size: int, gamma: float = 1e-1, verbose: bool = False):
        """
        :param optimizer:  _Optim - gradient descent optimizer.
        :param step_size: int - number of steps until decaying learning rate.
        :param gamma: float - learning rate decay factor.
        :param verbose: on/off logging every scheduler step.
        """
        super().__init__(optimizer=optimizer, verbose=verbose)
        self.step_size = step_size
        self.gamma = gamma
        self._epoch = 0

    def step(self):
        """
        Counts number of epochs and decays learning rate based on the current number of epochs and step_size.
        """
        self._epoch += 1

        if self._epoch % self.step_size == 0:
            self.optimizer.lr *= self.gamma
            if self.verbose:
                print(f"Epoch #{self._epoch} learning rate updated to {self.optimizer.lr}")


class ReduceLROnPlateau(_Scheduler):
    """
    Decays learning rate of optimizer when a metric has stopped improvement.
    """

    def __init__(self, optimizer: _Optim, factor: float = 1e-1, patience: int = 5, verbose: bool = False,
                 threshold: float = 1e-4):
        """
        :param optimizer: _Optim - gradient descent optimizer.
        :param factor: float - learning rate decay factor.
        :param patience: int - number of iterations without improvement.
        :param verbose: bool - on/off logging every scheduler step.
        :param threshold: float - measurement to focus only on significant improvements.
        """
        super().__init__(optimizer=optimizer, verbose=verbose)
        self.factor = factor
        self.patience = patience
        self.threshold = threshold
        self.wait = 0
        self.best_loss = np.inf

    def step(self, loss: float):
        """
        Increases self.wait by 1 if error is not improving and decays learning rate.
        of the optimizer of self.wait >= self.patience.

        :param loss: current values of the model error.
        """
        if loss < self.best_loss - self.threshold:
            self.best_loss = loss
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.optimizer.lr *= self.factor

                if self.verbose:
                    print(f"Learning rate updated to {self.optimizer.lr}")
                self.wait = 0



from abc import ABC

import numpy as np

from torchy.nn.values import Value


class Module(ABC):
    """
    Abstract class that represents any layer behavior and also used for type-hinting.
    """

    def __init__(self):
        self._train = True
        self._params = {}

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes gradient with respect to input.

        :param d_out: numpy array (n-dimensional) - gradient of loss with respect to output.
        :return: numpy array (n-dimensional) - gradient with respect to input.
        """
        pass

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Interface method that computes forward pass for layer.

        :param x: numpy array (n-dimensional) - previous forward pass.
        :return: numpy array(n-dimensional) - forward pass for this layer.
        """
        pass

    def __call__(self, x: np.ndarray, *args) -> np.ndarray:
        return self.forward(x, *args)

    def eval(self):
        """
        Set layer to evaluation mode
        """
        self._train = False

    def train(self):
        """
        Set layer to training mode
        """
        self._train = True

    @property
    def params(self) -> dict[str, Value]:
        """
        Collects all module parameters into dictionary.

        :return: dict[str, Value] - layer parameters.
        """
        if self._params is None:
            return {}

        if len(self._params) == 0:
            self._init_params()
            if len(self._params) == 0:
                self._params = None
                return {}

        return self._params

    def _init_params(self):
        """
        Goes through module attributes and sets it as parameter if it is an instance of Value.

        :return: None
        """
        for attr_name, attr_value in vars(self).items():
            if isinstance(attr_value, Value):
                self._params[attr_name] = attr_value



from setuptools import setup, find_packages

VERSION = '0.3.5.3'
DESCRIPTION = 'NumPy based neural network package with PyTorch-like API'

with open("README.md", "r") as fn:
    long_description = fn.read()

setup(
    name="torchy-nn",
    version=VERSION,
    author="Valentin Belyaev",
    author_email="chuvalik.work@gmail.com",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    url="https://github.com/chuvalniy/Torchy",
    install_requires=['numpy', 'pytest'],
    license="MIT",
    keywords=['python', 'neural net', 'from scratch', 'numpy', 'pytorch-like', 'cnn', 'dense'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



import numpy as np

from .module import Module
from torchy.nn.values import Value
from torchy.nn.initializations import init


class Conv2d(Module):
    """
    Two-dimensional convolutional neural network layer
    """

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int,
            stride: int = 1,
            padding: int = 0,
            bias: bool = True
    ):
        """
        :param in_channels: int - number of input channels.
        :param out_channels: int - number of output channels after convolution.
        :param kernel_size: int - size of convolutional kernel.
        :param stride: int - stride of convolutional kernel (default = 1).
        :param padding: int - padding added to all axes with respect to input (default = 0).
        :param bias: bool - consider bias in layer computation or not
        """
        super(Conv2d, self).__init__()

        self.weight: Value = init.kaiming_uniform(shape=(out_channels, in_channels, kernel_size, kernel_size))
        self.bias: Value | None = init.kaiming_uniform(shape=(out_channels,)) if bias else None
        self.x = None

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

        self._padding_width = ((0, 0), (0, 0), (padding, padding), (padding, padding))

    def forward(self, x: np.ndarray, *args) -> np.ndarray:
        """
        Computes forward pass for convolutional layer.

        :param x: numpy array (batch_size, in_channels, height, width) - incoming data.
        :return: numpy array (batch_size, out_channels, out_height, out_width) - incoming data after
        performing convolution operation on it.
        """
        self.x = x
        x_padded = np.pad(x, pad_width=self._padding_width)

        batch_size, in_channels, height, width = x.shape

        height += 2 * self.padding
        width += 2 * self.padding

        out_height = 1 + (height - self.kernel_size) // self.stride
        out_width = 1 + (width - self.kernel_size) // self.stride

        w_flattened = np.transpose(self.weight.data, axes=(2, 3, 1, 0)).reshape((-1, self.out_channels))

        out = np.zeros(shape=(batch_size, self.out_channels, out_height, out_width))
        for oh in range(out_height):
            for ow in range(out_width):
                oh_step = self.stride * oh
                ow_step = self.stride * ow

                input_region = x_padded[:, :, oh_step:oh_step + self.kernel_size, ow_step:ow_step + self.kernel_size]
                input_region_flattened = input_region.transpose((0, 2, 3, 1)).reshape((batch_size, -1))

                out[:, :, oh, ow] = np.dot(input_region_flattened, w_flattened)
                if self.bias is not None:
                    out[:, :, oh, ow] += self.bias.data

        return out

    def backward(self, d_out: np.ndarray) -> np.ndarray:
        """
        Computes gradient for convolutional layer with respect to input, weight and bias and also removes padding
        from the input layer that was added in the forward pass.

        :param d_out: numpy array (batch_size, out_channels, out_height, out_width) - gradient of loss function with
        respect to output of forward pass.
        :return: numpy array (batch_size, in_channels, height, width) - gradient with respect to input.
        """
        batch_size, in_channels, height, width = self.x.shape
        _, out_channels, out_height, out_width = d_out.shape

        x_padded = np.pad(self.x, pad_width=self._padding_width)

        w_flattened = self.weight.data.transpose((2, 3, 1, 0)).reshape((-1, self.out_channels))

        self.weight.grad = np.zeros(shape=(self.kernel_size, self.kernel_size, in_channels, out_channels))
        dw_flattened = self.weight.grad.reshape((-1, out_channels))

        d_x = np.zeros_like(x_padded)
        for oh in range(out_height):
            for ow in range(out_width):
                oh_step = self.stride * oh
                ow_step = self.stride * ow

                input_region = x_padded[:, :, oh_step:oh_step + self.kernel_size, ow_step:ow_step + self.kernel_size]
                input_region_flattened = input_region.transpose((0, 2, 3, 1)).reshape((batch_size, -1))
                d_out_pixel = d_out[:, :, oh, ow]

                dw_flattened += np.dot(input_region_flattened.T, d_out_pixel)
                dx_region_flattened = np.dot(d_out_pixel, w_flattened.T)
                dx_region = np.transpose(
                    dx_region_flattened.reshape((batch_size, self.kernel_size, self.kernel_size, in_channels)),
                    axes=(0, 3, 1, 2)
                )

                d_x[:, :, oh_step:oh_step + self.kernel_size, ow_step: ow_step + self.kernel_size] += dx_region

        if self.bias is not None:
            self.bias.grad = np.sum(d_out, axis=(0, 2, 3))

        self.weight.grad = self.weight.grad.transpose((3, 2, 0, 1))

        return d_x[:, :, self.padding:height + self.padding, self.padding:width + self.padding]



from typing import Sized

import numpy as np


# TODO: tests, docstring

class Dataset:
    def __getitem__(self, idx: int):
        raise NotImplemented("Subclasses of Dataset should implement __getitem__")

    def __len__(self):
        raise NotImplemented("Subclasses of Dataset should implement __len__")


class Sampler:
    def __iter__(self):
        NotImplemented("Subclasses of Sampler should implement __iter__")


class RandomSampler(Sampler):
    def __init__(self, data_source: Sized):
        self.data_source = data_source

    def __iter__(self):
        data_len = len(self.data_source)

        yield from np.random.permutation(data_len)

    def __len__(self):
        return len(self.data_source)


class SequentialSampler(Sampler):
    def __init__(self, data_source: Sized):
        self.data_source = data_source

    def __iter__(self):
        return iter(range(len(self.data_source)))

    def __len__(self):
        return len(self.data_source)


class DataLoader:
    def __init__(self, dataset: Dataset, batch_size: int = 32, sampler: Sampler = None):
        self.dataset = dataset
        self.batch_size = batch_size
        self.sampler = sampler if sampler is not None else SequentialSampler(list(range(len(dataset))))

        self._curr_index = 0

    def __iter__(self) -> "DataLoader":
        return self

    # TODO: add sampler logic to __next__
    def __next__(self):
        if self._curr_index >= len(self.dataset):
            raise StopIteration

        batch = self.dataset[self._curr_index:self._curr_index + self.batch_size]
        self._curr_index += self.batch_size

        return batch
